/*
Copyright (c) 2007 Takahiro Shinozaki

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

/*
  cmbnk
  Sample K integers from [1 N]

  staka
  Fri Jun 29 18:39:21 JST 2007
 */

#include <iostream>
#include <deque>
#include <algorithm>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include <limits.h>

using namespace std;
int verbose=0;

typedef deque<int> IVector;
typedef deque<IVector> IMatrix;

class combination {
public:
  static int cmb(unsigned N, unsigned K);
  static unsigned m_rand(unsigned N);
  static void listcombinations(unsigned N, unsigned K);
  static IVector idx2cmb(unsigned N, unsigned K, unsigned idx);
  /* Sampling functions */
  static IVector SampIntsAWOrep(unsigned m, unsigned n);
  static IMatrix mCnk(unsigned N, unsigned K, unsigned m);
  static IMatrix mCnkwoI(unsigned N, unsigned K, unsigned I, unsigned m);
  static IMatrix mSWRnk(unsigned N, unsigned K, unsigned m);
  static IMatrix mSWRnkwoI(unsigned N, unsigned K, unsigned I, unsigned m);
};


/* compute C(N,K) */
int
combination::cmb(unsigned N, unsigned K)
{
  unsigned i, j;
  IVector n, d;
  int c;
  int orgK;
  
  if (K==0) return 1;

  orgK = K;
  if (N-K < K) {
    K = N-K;
  }

  /* set numerator and denominator */
  for (i=0; i<K; i++) {
    n.push_back(N-i);
  }
  for (i=0; i<K; i++) {
    d.push_back(K-i);
  }

  // reduce
  for (i=0; i<K; i++) {
    for (j=0; j<K; j++) {
      for (c=d[j]; 0<c; c--) {
	if (!(n[i]%c) && !(d[j]%c)) {
          n[i] /= c;
          d[j] /= c;
        }
      }
    }
  }

  /* compute product */
  c = 1;
  for (i=0; i<K; i++) {
    if (INT_MAX/(n[i]+1) < c) {
      fprintf(stderr, "Error: combination C(N,K) overflow N=%d K=%d\n", N, orgK);
      exit(1);
    }
    c *= n[i];
  }
  return c;
}

/* return an integer uniformly sampled from [0 N-1] */
unsigned
combination::m_rand(unsigned N)
{
  unsigned retv;
  retv = (unsigned)ceil((float(rand())/RAND_MAX) * N);
  if (retv == 0) retv = 1; // this probability is small enough.
  retv--; // [0 N-1]

  return retv;
}

/* print all possible combinations. (This means it defines an order of combinations) */
void
combination::listcombinations(unsigned N, unsigned K)
{
  unsigned i, j;
  IVector cmb;
  bool eflg;

  if (K==0) return;

  /* Initialize */
  cmb.resize(K);
  for (i=0; i<K; i++) {
    cmb[i] = i;
  }
  cmb[K-1] = cmb[K-1]-1;
 
  /* list all the combinations */
  j = 0; eflg = false;
  while (1) { 
    i=K-1;
    //    while (0<=i) {
    while (1) {
      if (cmb[i]<int(N-(K-i))) {
	break;
      }
      if (i==0) {
	eflg = true;
	break;
      }
      i--;
    }
    if (eflg) {
      break;
    } else {
      cmb[i] = cmb[i] + 1;
      for (i++; i<K; i++) {
	cmb[i] = cmb[i-1] + 1;
      }
    }
    printf("cmb%u:", j++);
    for (i=0; i<K; i++) {
      printf(" %d", cmb[i]);
    }
    printf("\n");
  }
}

/* Get idx-th element from the combination list defined
   by listcombinations(N, K) */
IVector
combination::idx2cmb(unsigned N, unsigned K, unsigned idx)
{
  unsigned i, j;
  unsigned c, sum;
  IVector comb;

  comb.resize(K);
  //  if (idx < 0 || cmb(N, K) <= (int)idx) {
  if (cmb(N, K) <= (int)idx) {
    fprintf(stderr, "idx is out of range idx %d [%d %d]\n", (int)idx, 0, cmb(N, K)-1);
    exit(1);
  }
  j = 0;
  for (i=0; i<K; i++) {
    sum = 0;
    for (; j<N-K+i; j++) { // N-K+i is the max value at i-th element
      c = cmb(N-j-1, K-i-1);
      if (idx < sum + c) {
	break; 
      }
      sum = sum + c;
    }
    comb[i] = j;
    idx = idx - sum;
    j++;
  }
  return comb;
}

/* Sample integers from [0 m-1] for n times approximately without replacement */
IVector
combination::SampIntsAWOrep(unsigned m, unsigned n)
{
  IVector tmp, rnd;
  unsigned i;

  if (m < n) {
    fprintf(stderr, "ERROR SampIntsAWOrep: n must be smaller than m. m=%d n=%d.\n", m, n);
    exit(1);
  }

  rnd.resize(n);
  if (m<100000 || m/50 < n) {
    // strict sampling without replacement
    tmp.resize(m);
    for (i=0; i<m; i++) tmp[i] = i;
    for (i=0; i<10; i++) {
      // for better randomness (Though it still looks not very good)
      random_shuffle(tmp.begin(), tmp.end(), m_rand); 
    }
    /* #ifdef HPC */
    // avoid to use random_sample.
    // tmp must be shuffled well.
    for (i=0; i<n; i++) {
      rnd[i] = tmp[i];
    }
    /* #else
    random_sample(tmp.begin(), tmp.end(), rnd.begin(), rnd.end(), m_rand);
    #endif
    */
  } else {
    // approximation by sampling with replacement
    for (i=0; i<n; i++) {
      rnd[i] = m_rand(m);
    }
  }
  return rnd;
}

/* sample C(N,K) for m times (approximately) without replacement. */
IMatrix
combination::mCnk(unsigned N, unsigned K, unsigned m)
{
  unsigned numc;
  IVector samples, comb;
  unsigned i, idx;
  IMatrix retv;

  retv.resize(m);
  numc = cmb(N, K);
  samples = SampIntsAWOrep(numc, m);
  for (i=0; i<m; i++) {
    idx = samples[i];
    comb = idx2cmb(N, K, idx);
    retv[i] = comb;
  }
  return retv;
}

/* sample C(N,K) that does not include I for m times 
   (approximately) without replacement. */
IMatrix
combination::mCnkwoI(unsigned N, unsigned K, unsigned I, unsigned m)
{
  IMatrix retv;
  unsigned i, j;

  retv = mCnk(N-1, K, m);
  for (i=0; i<m; i++) {
    for (j=0; j<K; j++) {
      if ((int)I <= retv[i][j]) {
	retv[i][j]++;
      }
    }
  }
  return retv;
}

/* do sample with replacement for m times */
IMatrix 
combination::mSWRnk(unsigned N, unsigned K, unsigned m)
{
  IMatrix retv;
  IVector swr;
  unsigned i, j;

  retv.resize(m);
  for (i=0; i<m; i++) {
    swr.resize(K);
    for (j=0; j<K; j++) {
      swr[j] = (int)m_rand(N); /* [0 N-1] */
    }
    retv[i] = swr;
  }
  return retv;
}

/* sample with replacement that does not include I for m times 
   (approximately) without replacement. */
IMatrix
combination::mSWRnkwoI(unsigned N, unsigned K, unsigned I, unsigned m)
{
  IMatrix retv;
  unsigned i, j;

  retv = mSWRnk(N-1, K, m);
  for (i=0; i<m; i++) {
    for (j=0; j<K; j++) {
      if ((int)I <= retv[i][j]) {
	retv[i][j]++;
      }
    }
  }
  return retv;
}

void
version(void)
{
  cout << "nbmm ver 1.1\n";
}

void
usage(char *myname)
{
  version();
  cout << myname << " [options] N K\n";
  cout << "sample K numberes from [1 N]\n";
  cout << "  options:\n";
  cout << "	-s seed\n";
  cout << "	-e exclude\n";
  cout << "	-m sample_m_sets\n";
  cout << "	-i print_idx\n";
  cout << "	-r sample_with_replacement\n";
  cout << "	-v verbose_level\n";
}
/* 
   Note: internary, array's indices starts with 0.
   However, as a unix command, this program counts
   integer from 1.
*/

int
main(int argc, char **argv)
{
  int seed, exclude, n, k, m, idx;
  int i, j;
  int c;
  IMatrix rstbl;
  int stype;

  stype = 0; // without replacement
  idx = -1;
  seed = 1;
  exclude = 0;
  m = 1;
  while (1) {
    c = getopt(argc, argv, "s:e:m:i:rv:");
    if (c == -1)
      break;
    switch (c) {
    case 's':
      seed = atoi(optarg);
      srand(seed);
      break;
    case 'e':
      exclude = atoi(optarg);
      break;
    case 'm':
      m = atoi(optarg);
      break;
    case 'i':
      idx = atoi(optarg)-1;
      break;
    case 'r':
      stype = 1; // with replacement
      break;
    case 'v':
      verbose = atoi(optarg);
      break;
    default:
      break;
    }
  }
  if (argc - optind != 2) {
    usage(argv[0]);
    exit(1);
  } else {
    n = atoi(argv[optind++]);
    k = atoi(argv[optind++]);
  }
  if (0<verbose) {
    printf("n=%d k=%d seed=%d ex=%d\n", n, k, seed, exclude);
  }
  if (stype == 0) {
    if (0 <= (exclude-1)) {
      rstbl = combination::mCnkwoI(n, k, exclude-1, m);
    } else {
      rstbl = combination::mCnk(n, k, m);
    }
  } else {
    if (0<=exclude-1) {
      rstbl = combination::mSWRnkwoI(n, k, exclude-1, m);
    } else {
      rstbl = combination::mSWRnk(n, k, m);
    }
  }
  if (0<=idx) {
    for (j=0; j<k; j++) {
      cout << rstbl[idx][j]+1 << " ";
    }
    cout << endl;
  } else {
    for (i=0; i<m; i++) {
      for (j=0; j<k; j++) {
	cout << rstbl[i][j]+1 << " ";
      }
      cout << endl;
    }
  }
  //  combination::listcombinations(n,k);
  return 0;
}
