#! /bin/bash -e

# Copyright (c) 2007 Takahiro Shinozaki

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Synopsis
#   agem.bash [-debug] [-config config] [-doneflg doneflg] [-envlog envlog]
#
# Description:
#	Update HMM model parameters by the EM method

# Change log:
# Tue May 13 13:05:33 JST 2008
# A bug fix in writing stats. Use $outdir/en$en/stats instead of $outdir/stats

# Sat May 10 22:21:48 JST 2008
# A bug fix in brun config selection.
# Move $err to $err.bak.$retry in retry as $out.
# By Takahiro Shinozaki.

# Fri Mar 14 18:57:16 JST 2008
# The initial version was written by Takahiro Shinozaki.


[ -n "$debug" ] && set -x
scriptp=$0
script=`basename $0`

function usage {
    echo "Usage: $script [-debug]"
    echo "          [-config CONFIGURATION]"
    echo "          [-doneflg DONE_FLAG]"
    echo "          [-envlog ENV_LOG]"
    exit 1
}

# initialize local variables
doneflg=done
envlog=env.log

while [ -n "$1" ]; do
    case "$1" in
	-debug) shift; export debug=1; set -x;;
	-config) shift; export config=$1; shift; . $config;;
	-doneflg) shift; doneflg="$1"; shift;;
	-envlog) shift; envlog="$1"; shift;;
	*) usage;
    esac
done

if [ -f $doneflg ]; then
    echo Nothing to be done.
    exit 0
fi

# dump environmental variables
env > $envlog

# print process info
echo -----
date
pwd
echo $script
# All the input global variables directly used in this script are (should be) listed here.
echo debug=$debug
echo config=$config
echo HEREst=$HERest
echo accss=$accss
echo cmbnk=$cmbnk
echo mlf=$mlf
echo nparallel=$nparallel
echo sdscpd=$sdscpd
echo hmm0=$hmm0
echo hlist=$hlist
echo niter=$niter
echo beam=$beam
echo nensemble=$nensemble
echo nselect=$nselect
echo brun=$brun
echo bruncfg=$bruncfg
echo bruncfgL=$bruncfgL
echo bruncfgD=$bruncfgD
echo brunLsize=$brunLsize
echo rmiaccs=$rmiaccs

# check requirments
if [ "$HERest" = "" ]; then echo HERest is not defined; exit 1; fi
if [ "$accss" = "" ]; then echo accss is not defined; exit 1; fi
if [ "$cmbnk" = "" ]; then echo cmbnk is not defined; exit 1; fi
if [ "$mlf" = "" ]; then echo mlf is not defined; exit 1; fi
if [ "$nparallel" = "" ]; then echo nparallel is not defined; exit 1; fi
if [ "$sdscpd" = "" ]; then echo sdscpd is not defined; exit 1; fi
if [ "$hmm0" = "" ]; then echo hmm0 is not defined; exit 1; fi
if [ "$hlist" = "" ]; then echo hlist is not defined; exit 1; fi
if [ "$niter" = "" ]; then echo niter is not defined; exit 1; fi
if [ "$beam" = "" ]; then echo beam is not defined; exit 1; fi
if [ "$nensemble" = "" ]; then echo nensemble is not defined; exit 1; fi
if [ "$nselect" = "" ]; then echo nselect is not defined; exit 1; fi
if [ "$brun" = "" ]; then echo brun is not defined; exit 1; fi


for iter in `seq 1 $niter`; do
    piter=$(( $iter -1 ))
    outdir=iter-$iter
    if [ -f $outdir/done ]; then
	echo skip iter-$iter.
	continue;
    fi
    mkdir -p $outdir

    # find a suitable execution configuration
    if [ $piter = 0 ]; then
	srchmm=$hmm0
    else
	srchmm=iter-$piter/hmmdefs
    fi
    if [ -n "$bruncfg" -a -n "$bruncfgL" -a -n "$brunLsize" ]; then
	msize=`du $srchmm | cut -f 1`
	if [ $brunLsize -lt $msize ]; then
	    pembruncfg=$bruncfgL
	else
	    pembruncfg=$bruncfg
	fi
    elif [ -n "$bruncfg" ]; then
	pembruncfg=$bruncfg
    else 
	pembruncfg=
    fi
    if [ -n "$pembruncfg" ]; then
	brunopt="-c $pembruncfg"
    else
	brunopt=""
    fi

    # E-step
    # parallel processing
    (
    for pp in `seq 1 $nparallel`; do
	mkdir -p $outdir/lot$pp
	scp=$sdscpd/scp.$pp
        # process the same data with different ensemble models
	for en in `seq 1 $nensemble`; do
	    if [ $piter = 0 ]; then
	         srchmm=$hmm0
	    else
		 srchmm=iter-$piter/en$en/hmmdefs
	    fi
	    out=$outdir/estep.lot$pp.en$en.out
	    err=$outdir/estep.lot$pp.en$en.err
	    cmd="time $HERest -T 1 -p $en -I $mlf -S $scp -H $srchmm -M $outdir/lot$pp -t $beam $hlist"
	    echo "$out $err $cmd"
        done
    done
    ) > $outdir/estep-brun
    $brun $brunopt $outdir/estep-brun

    # check and retry
    retry=0
    while [ 1 ]; do
	retry=$((retry+1))
	try=$((retry-1))
	nfail=0
	rm -f $outdir/estep-brun.retry.$retry
	for pp in `seq 1 $nparallel`; do
	    nut=`cat $sdscpd/scp.$pp | wc -l`
	    for en in `seq 1 $nensemble`; do
		out=$outdir/estep.lot$pp.en$en.out
		err=$outdir/estep.lot$pp.en$en.err
		npu=`grep Utterance $out | wc -l`
		ntry=5
		while [ $ntry -gt 0 ] && [ $nut -ne $npu ]; do
                # Try to wait. Sometimes, it takes a while before the file on NFS is updated. 
		    sleep 5
		    npu=`grep Utterance $out | wc -l`
		    ntry=$((ntry-1))
		done
		if [ $nut -ne $npu ]; then
		    echo "WARN: failed to process some utterances: $out $npu/$nut (try=$try)"
		fi
		if [ ! -s $outdir/lot$pp/HER$en.ss ]; then
		# Retry running HERest (Sometimes it works on some environments such as HPC2500)
		    echo Retry HERest for lot$pp en$en
		    nfail=$((nfail+1))
		    scp=$sdscpd/scp.$pp
		    if [ $piter = 0 ]; then
			srchmm=$hmm0
		    else
			srchmm=iter-$piter/en$en/hmmdefs
		    fi
		    if [ -e $out ]; then
			mv $out $out.bak.$retry
		    fi
		    if [ -e $err ]; then
			mv $err $err.bak.$retry
		    fi
		    cmd="time $HERest -T 1 -p $en -I $mlf -S $scp -H $srchmm -M $outdir/lot$pp -t $beam $hlist"
		    echo "$out $err $cmd" >> $outdir/estep-brun.retry.$retry
		fi
	    done
	done
	if [ $nfail -gt 0 ]; then
	    if [ $retry -gt 3 ]; then
		echo Abort: retry failed
		exit 1
	    else
		$brun $brunopt $outdir/estep-brun.retry.$retry
		sleep 30
	    fi
	else
	    break
	fi
    done

    # compute averaged sufficient statistics for each lot
    (
    for pp in `seq 1 $nparallel`; do
	# This part is disk intensive
	out=$outdir/accss$pp.out
	err=$outdir/accss$pp.err
	cmd="time $accss -n $outdir/HER$pp.ss $outdir/lot$pp/HER*.ss"
	echo "$out $err $cmd"
    done
    ) > $outdir/ssavr-brun
    $brun $brunopt $outdir/ssavr-brun
    nhss=`ls $outdir/HER*.ss | wc -l`
    if [ $nhss -ne $nparallel ]; then
	echo Abort: accss failed to compute averaged sufficient statistics
	exit 1
    fi
    # clean the temporary SS files
    for pp in `seq 1 $nparallel`; do
	rm $outdir/lot$pp/HER*.ss
    done

    # M-step
    if [ $piter = 0 ]; then
	srchmm=$hmm0
    else
	srchmm=iter-$piter/hmmdefs
    fi
    # accumulate sufficient statistics for general model
    out=$outdir/mstep.out
    err=$outdir/mstep.err
    cmd="time $HERest -T 1 -p 0 -s $outdir/stats -H $srchmm -M $outdir $hlist $outdir/HER*.ss"
    echo $out $err $cmd | $brun $brunopt -
    # accumulate sufficient statistics for ensemble models
    (
    for en in `seq 1 $nensemble`; do
	mkdir -p $outdir/en$en
	out=$outdir/mstep.en$en.out
	err=$outdir/mstep.en$en.err
	seed=$((en*iter))
	acc=""
	for i in `$cmbnk -s $seed -m $nensemble -i $en $nparallel $nselect`; do
	    acc="$acc $outdir/HER$i.ss"
	done
	cmd="time $HERest -T 1 -p 0 -s $outdir/en$en/stats -H $srchmm -M $outdir/en$en $hlist $acc"
	echo "$out $err $cmd"
    done
    ) > $outdir/mstep-brun
    $brun $brunopt $outdir/mstep-brun

    if [ "$rmiaccs" = "true" ]; then
	if [ $iter -ne $niter ]; then
	    rm -f $outdir/HER*.ss
	fi
    fi
    touch $outdir/done
done

touch $doneflg
echo done $script
