#! /bin/bash -e

# Copyright (c) 2007 Takahiro Shinozaki

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Synopsis
#   cvem.bash [-debug] [-config config] [-doneflg doneflg] [-envlog envlog]
#
# Description:
#	Update HMM model parameters by the CV-EM method

# Change log:
# Mon May 12 13:44:56 JST 2008
# The initial version was written by Takahiro Shinozaki.

[ -n "$debug" ] && set -x
script=`basename $0`

function usage {
    echo "Usage: $script [-debug]"
    echo "          [-config CONFIGURATION]"
    echo "          [-doneflg DONE_FLAG]"
    echo "          [-envlog ENV_LOG]"
    exit 1
}

# initialize local variables
doneflg=done
envlog=env.log

while [ -n "$1" ]; do
    case "$1" in
	-debug) shift; export debug=1; set -x;;
	-config) shift; export config=$1; shift; . $config;;
	-doneflg) shift; doneflg="$1"; shift;;
	-envlog) shift; envlog="$1"; shift;;
	*) usage;
    esac
done

if [ -f $doneflg ]; then
    echo Nothing to be done.
    exit 0
fi

# dump environmental variables
env > $envlog

# print process info
echo -----
date
pwd
echo $script
# All the input global variables directly used in this script are (should be) listed here.
echo debug=$debug
echo config=$config
echo HEREst=$HERest
echo mlf=$mlf
echo nparallel=$nparallel
echo sdscpd=$sdscpd
echo hmm0=$hmm0
echo hlist=$hlist
echo niter=$niter
echo beam=$beam
echo brun=$brun
echo bruncfg=$bruncfg
echo bruncfgL=$bruncfgL
echo bruncfgD=$bruncfgD
echo brunLsize=$brunLsize
echo rmiaccs=$rmiaccs

# check requirments
if [ "$HERest" = "" ]; then echo HERest is not defined; exit 1; fi
if [ "$accss" = "" ]; then echo accss is not defined; exit 1; fi
if [ "$mlf" = "" ]; then echo mlf is not defined; exit 1; fi
if [ "$nparallel" = "" ]; then echo nparallel is not defined; exit 1; fi
if [ "$sdscpd" = "" ]; then echo sdscpd is not defined; exit 1; fi
if [ "$hmm0" = "" ]; then echo hmm0 is not defined; exit 1; fi
if [ "$hlist" = "" ]; then echo hlist is not defined; exit 1; fi
if [ "$niter" = "" ]; then echo niter is not defined; exit 1; fi
if [ "$beam" = "" ]; then echo beam is not defined; exit 1; fi
if [ "$brun" = "" ]; then echo brun is not defined; exit 1; fi


for iter in `seq 1 $niter`; do
    piter=$(( $iter -1 ))
    outdir=iter-$iter
    if [ -f $outdir/done ]; then
	echo skip iter-$iter.
	continue;
    fi
    mkdir -p $outdir

    # find a suitable execution configuration
    if [ $piter = 0 ]; then
	srchmm=$hmm0
    else
	srchmm=iter-$piter/hmmdefs
    fi
    if [ -n "$bruncfg" -a -n "$bruncfgL" -a -n "$brunLsize" ]; then
	msize=`du $srchmm | cut -f 1`
	if [ $brunLsize -lt $msize ]; then
	    pembruncfg=$bruncfgL
	else
	    pembruncfg=$bruncfg
	fi
    elif [ -n "$bruncfg" ]; then
	pembruncfg=$bruncfg
    else 
	pembruncfg=
    fi
    if [ -n "$pembruncfg" ]; then
	brunopt="-c $pembruncfg"
    else
	brunopt=""
    fi

    # E-step
    # parallel processing
    (
    for pp in `seq 1 $nparallel`; do
	if [ $piter = 0 ]; then
	    srchmm=$hmm0
	else
	    srchmm=iter-$piter/lot$pp/hmmdefs
	fi
	scp=$sdscpd/scp.$pp
	out=$outdir/estep.$pp.out
	err=$outdir/estep.$pp.err
	cmd="time $HERest -T 1 -p $pp -I $mlf -S $scp -H $srchmm -M $outdir -t $beam $hlist"
	echo "$out $err $cmd"
    done
    ) > $outdir/estep-brun
    $brun $brunopt $outdir/estep-brun

    # check and retry
    retry=0
    while [ 1 ]; do
	retry=$((retry+1))
	try=$((retry-1))
	nfail=0
	rm -f $outdir/estep-brun.retry.$retry
	for pp in `seq 1 $nparallel`; do
	    nut=`cat $sdscpd/scp.$pp | wc -l`
	    out=$outdir/estep.$pp.out
	    err=$outdir/estep.$pp.err
	    npu=`grep Utterance $out | wc -l`
	    ntry=5
	    while [ $ntry -gt 0 ] && [ $nut -ne $npu ]; do
                # Try to wait. Sometimes, it takes a while before the file on NFS is updated. 
		sleep 5
		npu=`grep Utterance $out | wc -l`
		ntry=$((ntry-1))
	    done
	    if [ $nut -ne $npu ]; then
		echo "WARN: failed to process some utterances: $out $npu/$nut (try=$try)"
	    fi
	    if [ ! -s $outdir/HER$pp.ss ]; then
		echo Retry HERest for lot$pp
		nfail=$((nfail+1))
		scp=$sdscpd/scp.$pp
		if [ $piter = 0 ]; then
		    srchmm=$hmm0
		else
		    srchmm=iter-$piter/lot$pp/hmmdefs
		fi
		if [ -e $out ]; then
		    mv $out $out.bak.$retry
		fi
		if [ -e $err ]; then
		    mv $err $err.bak.$retry
		fi
		cmd="time $HERest -T 1 -p $pp -I $mlf -S $scp -H $srchmm -M $outdir -t $beam $hlist"
		echo "$out $err $cmd" >> $outdir/estep-brun.retry.$retry
	    fi
	done
	if [ $nfail -gt 0 ]; then
	    if [ $retry -gt 3 ]; then
		echo Abort: retry failed
		exit 1
	    else
		$brun $brunopt $outdir/estep-brun.retry.$retry
		sleep 30
	    fi
	else
	    break
	fi
    done

    # M-step
    if [ -n "$bruncfgD" ]; then
	brunopt="-c $bruncfgD"
    else
	brunopt=""
    fi
    if [ $piter = 0 ]; then
	srchmm=$hmm0
    else
	srchmm=iter-$piter/hmmdefs
    fi

    # accumulate all the sufficient statistics files
    out=$outdir/accss.out
    err=$outdir/accss.err
    if [ -e $outdir/HER.ss ]; then
	rm $outdir/HER.ss
    fi
    cmd="time $accss -s $outdir/HER.ss $outdir/HER*.ss"
    echo $out $err $cmd | $brun $brunopt -

    # update model parameters of a general model
    out=$outdir/mstep.out
    err=$outdir/mstep.err
    cmd="time $HERest -T 1 -p 0 -s $outdir/stats -H $srchmm -M $outdir $hlist $outdir/HER.ss"
    echo $out $err $cmd | $brun $brunopt -
    # update model parameters of CV models
    (
    for pp in `seq 1 $nparallel`; do
	mkdir -p $outdir/lot$pp
	out=$outdir/lot$pp/mstep.out
	err=$outdir/lot$pp/mstep.err
	# When updating CV models, i-th SS is subtracted from the general SS
	cmd="time $HERest -T 1 -p 0 -s $outdir/lot$pp/stats -H $srchmm -M $outdir/lot$pp \
                     $hlist $outdir/HER.ss minusSS $outdir/HER$pp.ss"
	echo "$out $err $cmd"
    done
    ) > $outdir/mstep-brun
    $brun $brunopt $outdir/mstep-brun

    for pp in `seq 1 $nparallel`; do
	if [ ! -e $outdir/lot$pp/hmmdefs ]; then
	    echo Abort: failed to make a hmm : $outdir/lot$pp
	    exit 1
	fi
    done
    if [ "$rmiaccs" = "true" ]; then
	if [ $iter -ne $niter ]; then
	    rm -f $outdir/HER*.ss
	fi
    fi
    touch $outdir/done
done

touch $doneflg
echo done $script
