#! /usr/bin/perl


# Copyright (c) 2008 Takahiro Shinozaki
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


# Synopsis
#   mpbsrun.pl [-c config] brun
#
# Description:
#   Run multiple processes via PBS according to "brun" input file.
#   Unless explicitly specified, the processes run at a directory
#   where this command is invoked.
#
#   -c config  : Specifies config file. Each line of the config is
#                a command option of the qsub command or a meta-
#                command. Lines begin with # are ignored.
#                The meta-commands are:
#                LOCAL : Run jobs locally without using a queue system
#                MAXRUN num : Limit # of concurrent jobs to "num"
#   brun       : Specifies job definition file. If '-' is specified, 
#                standard input is read instead of a file.
#
# mpbsrun.pl is a member of mXXrun scripts.
# mXXrun is a family of lapper scripts for job execution
# using various batch systems such as PBS, LSF, condor, etc
# or without using batch system.
# mXXrun should have the same command line syntax.
# mXXrun reads all the job submission information if any
# from the config file every time when it is invoked.
# While the config file may vary, brun file format should 
# be common over the mXXrun family scripts.
# mXXrun blocks until all the processes terminate.
#
# The format of the brun file is:
# -----
# std1.out std1.err command1 [command options] 
# std2.out std2.err command2 [command options] 
#                  |
# stdN.out stdN.err commandN [command options] 
# -----

# Change log:
# Sat May 10 19:42:24 JST 2008
# Added MAXRUN option for config file : by Takahiro Shinozaki.

# Wed May  7 19:58:11 JST 2008
# Use -N option for qusb : by Takahiro Shinozaki.

# Sun Dec 23 19:52:05 JST 2007
# The initial version was written by Takahiro Shinozaki.

require "getopts.pl";
&Getopts('c:');

$myname = $0;
$pwd = `pwd`; chomp($pwd);
$qoptions = "-d $pwd ";
%jids = ();
$localflg = 0;
$maxrun = 0;
$nqfinished = 0;
$pid = $$;

sub usage() {
    die "mpbsrun.pl [-c config] brun\n";
}

if ($opt_c) {
	&read_config($opt_c);
}

if (@ARGV != 1) {
	printf STDERR "%s : Invalid arguments: @ARGV\n", $myname;
	&usage();
}

# open command list file
if ($ARGV[0] eq '-') {
    $in = STDIN;
} else {
    open(IN, $ARGV[0]) || die "$myname : $! : $ARGV[0]\n";
    $in = IN;
}

# run jobs
$jnum=0;
$runjnum=0;
mkdir "brun.$pid";
while (<$in>) {
    next if (/^\#/);
    ($out, $err, $cmd, @opts) = split;
    $jnum++;
    $runjnum++;
#    print "DEBUG $maxrun $runjnum\n";
    if ((0 < $maxrun) && ($maxrun < $runjnum)) {
	&wait_a_job();
	$runjnum--;
    }
    $cmdf = "brun.$pid/job$jnum";
    open(CML, ">$cmdf") || die "$! : $cmdf\n";
    print CML "#! /bin/bash\n";
    print CML "# This file is automatically created by $0\n";
    print CML "hostname\n";
    print CML "$cmd @opts\n";
    close(CML);
    chmod(0755, $cmdf);

    if ($localflg) {
	if (fork() == 0) {
	    exec("$cmdf 1> $out 2> $err");
	}
    } else {
	$bcmd = "qsub -N brun.p$pid.j$jnum -o $out -e $err $qoptions $cmdf";
	$jid = `$bcmd`;
	$jid =~ /^(\d+)/; # this number is uniq
	if ($jid eq "") {
	    die "Failed to submit a job: $bcmd\n";
	}
	$jidn = $1;
	$jids{$jidn} = 1;
    }
}

# Wait all the jobs to terminate.
while (0<$runjnum) {
    &wait_a_job();
    $runjnum--;
}

system("rm brun.$pid/job*");
rmdir "brun.$pid";

sub wait_a_job() {
    my ($jid, $jidn, $name, $user, $time, $st, $qu, %jidstmp);

    if ($localflg) {
	wait;
    } else {
	if (0<$nqfinished) {
	    $nqfinished --;
	} else {
	    die "$myname : Assertion error 1\n" if ($nqfinished != 0);
	    # Since Torque (OpenPBS) does not support blocking with qsub,
	    # we need to monitor the jobs by ourself.
	    # FYI, qsub in PBSPro5.4 supports the blocking with -W block=true
	    # option.
	    while (1) {
		sleep 15; # May not be too short as it increases server load.
		open(QST, "qstat|") || die "$myname : $! qstat\n";
		$_ = <QST>;
		%jidstmp = ();
		if (/^$/) {
		    # No job is left
		} elsif (!/^Job id/) {
		    die "Unexpected format: $_\n";
		} else {
		    while (<QST>) {
			chomp;
			next if (/^-------------/);
			($jid, $name, $user, $time, $st, $qu) = split;
			$jid =~ /^(\d+)/;
			$jidn = $1;
			if ($st ne "C") {
			    # a list of jobs that has not been finished
			    $jidstmp{$jidn} = 1;
			}
		    }
		}
		close(QST);
		foreach $jidn (keys %jids) {
		    if (!$jidstmp{$jidn}) {
			# A job has finished.
#			print "DEBUG delete $jidn nqfinished $nqfinished\n";
			delete $jids{$jidn};
			$nqfinished ++;
		    }
		}
		if (0<$nqfinished) {
		    $nqfinished --;
		    last;
		}
	    } # while
	}
    }
}

sub read_config($) {
    my ($cfg) = @_;

    open(CFG, $cfg) || die "$myname : $! : $ARGV[0]\n";
    while (<CFG>) {
	chomp;
	if (/^\#/) {
	} elsif (/LOCAL/) {
	    $localflg = 1;
	} elsif (/MAXRUN/) {
	    if (/MAXRUN\s+(\d+)/) {
		$maxrun = $1;
		next;
	    } else {
		die "$myname : Format error in config file $cfg: MAXRUN num: $ARGV[0]\n";
	    }
	} else {
	    $qoptions .= "$_ ";
	}
    }
}
